/*
PROducer Firmware
VERSION 0.9
[c]2013 DCR Dirk Conrad

DDR[x]  ... Datenrichtung (0=Eingang; 1 = Ausgang)
PORT[x] ... Ausgabe auf Port oder 1= PullUp Widerstand bei Datenrichtung einlesen
PIN[x]  ... Einlesen vom Port
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <compat/ina90.h>
#include <stdio.h>
#include <util/twi.h>

// Def
#define MAX_FRAMES		2621
#define I2C_SLA_ADR		0x50

// Sub
void Init (void);
void Draw_Workspace (void);
void I2C_Init(void);
void I2C_Start (void);
void I2C_Stop (void);
void I2C_Adr_W (void);
void I2C_Adr_R (void);
void I2C_Data_W (unsigned char ucByte);
unsigned char I2C_Data_R (void);
unsigned char I2C_Data_R_Ack (void);
void Wait_TWINT_Flag (void);
void Wait_TWSTO_Flag (void);
unsigned char EE_Read (unsigned int iAdr);
void EE_Write (unsigned int iAdr);
void USART_Init (void);
void USART_Transmit (char *data);
void USART_Receive (void);
void USART_Exec_Commands (void);

// Var
volatile unsigned char ucWorkspace[2][3][8] = { 
	{ { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } },

	{ { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } }
	};
volatile unsigned char ucDark[2] = { 0, 0 };
unsigned char ucTransition = 0;
unsigned char ucSkipCounter = 0;
volatile unsigned char ucFirstSkipCounter = 0;
volatile unsigned int iPos = 0;
volatile unsigned char ucRowCnt = 0;
volatile unsigned char ucColCnt = 0;
volatile unsigned char ucFrame = 0;
volatile unsigned char ucColorCnt = 0;
volatile unsigned char ucColor = 0;
volatile unsigned char ucNewRow = 0;
volatile unsigned char ucDrawFrame = 1;
volatile unsigned char red[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
volatile unsigned char grn[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
volatile unsigned char ucPortA = 0;
volatile unsigned char ucPortB = 0;
volatile unsigned char ucErr = 0x00;
unsigned char ucEEWriteBuffer[128];
unsigned char sUSARTBuffer[136];
volatile int iUSARTBufferPointer = 0;
volatile unsigned int uiUSARTBusyCounter = 0;
volatile unsigned char ucMaster = 0;
volatile unsigned char ucLastSyncBit = 0;
volatile int iMissingSyncCounter = 0;
unsigned char ucFade;
unsigned char ucFadeIn;
unsigned char ucFadeOut;
volatile unsigned long ulWait;

// ########## Hauptprogramm ##########

int main (void) {
int iX;

	Init ();
	USART_Init ();
	I2C_Init ();

	ucTransition = EE_Read (24);
	ucFirstSkipCounter = (ucTransition & 0x3F);
	ucSkipCounter = ucFirstSkipCounter;

	// Enable Interrupts
	sei();

	while (1) {
		USART_Receive ();
		USART_Exec_Commands ();

		if (ucNewRow) {
			ucNewRow = 0;

			if (ucFadeIn == 0) {
				if (ucSkipCounter == 0) {
					if (ucFadeOut == 0) {
						iPos += 25;
						ucTransition = EE_Read (iPos+24);
						ucSkipCounter = (ucTransition & 0x3F);
						if (iPos == 65525 || ucSkipCounter == 63) {
							iPos = 0;
							ucTransition = EE_Read (iPos+24);
							ucSkipCounter = (ucTransition & 0x3F);
						}
						ucSkipCounter *= 3;
						ucFade = (ucTransition & 0xC0) >> 6; 
						switch (ucFade) {
							case 0: ucFadeIn =  0; ucFadeOut =  0; break;
							case 1: ucFadeIn = 16; ucFadeOut =  0; break;      
							case 2: ucFadeIn =  0; ucFadeOut = 16; break;      
							case 3: ucFadeIn = 16; ucFadeOut = 16; break; 
						}
					}
				}
				else {
					ucSkipCounter--;
				}
			}
			ucDark[ucDrawFrame] = 0;
			if (ucFadeIn > 0) {
				ucDark[ucDrawFrame] = ucFadeIn;
				ucFadeIn--;
			}
			else {
				if (ucFadeOut > 0 && ucSkipCounter == 0) {
					ucDark[ucDrawFrame] = 16 - ucFadeOut;
					ucFadeOut--;
				}
			}
		
			for (iX=0; iX<8; iX++) {
				ucWorkspace[ucDrawFrame][0][iX] = EE_Read (iPos+iX);
			}
			for (iX=0; iX<8; iX++) {
				ucWorkspace[ucDrawFrame][1][iX] = EE_Read (iPos+iX+8);
			}
			for (iX=0; iX<8; iX++) {
				ucWorkspace[ucDrawFrame][2][iX] = EE_Read (iPos+iX+16);
			}

			// Debug ||
			ucWorkspace[ucDrawFrame][0][0] |= ucErr;
		}

	}
	return 0;
}

// ########## Initialisierung ##########

void Init (void) {
	// Disable JTAG
	MCUCSR |= 0x80;
	MCUCSR |= 0x80;

	// Port A
	// RED (OUT)
	DDRA = 0xFF;
	PORTA = 0xFF;

	// Port B
	// GREEN (OUT)
	DDRB = 0xFF;
	PORTB = 0xFF;

	// Port C
	// 0x01 SCL
	// 0x02 SDA
	// 0x04 ROW A (OUT)
	// 0x08 ROW B (OUT)
	// 0x10 ROW C (OUT)
	// 0x20 SYNC IN (IN)
	// 0x40 SYNC OUT (OUT)
	// 0x80 MASTER/SLAVE (IN)
	DDRC = 0x5C;
	PORTC = 0xA0;

	// Port D
	// 0x01 RXD (IN)
	// 0x02 TXD (OUT)
	// 0x04 DATA IN 0 (IN)
	// 0x08 DATA IN 1 (IN)
	// 0x10 DATA IN 2 (IN)
	// 0x20 DATA OUT 0 (OUT)
	// 0x40 DATA OUT 1 (OUT)
	// 0x80 DATA OUT 2 (OUT)
	DDRD = 0xE2;
	PORTD = 0x1D;

	// MASTER/SLAVE Bestimmung (Wartezeit vor dem ersten Lesen)
	for (ulWait=0; ulWait < 30000; ulWait++) { _NOP (); }
	if (PINC & 0x80) {
		ucMaster = 1;
		for (ulWait=0; ulWait < 300000; ulWait++) {
			_NOP ();
		}
	}

	// Init Timer 1
	TCCR1A |= 0x00;
	TCCR1B |= ((1<<WGM12)|(0<<CS12)|(1<<CS11)|(1<<CS10));	// Prescaler 64 und..
	if (ucMaster == 1) { OCR1A = 39; }							// Output Compare Register A (16 MHz : 64 : 39 = 6410 kHz)
	if (ucMaster == 0) { OCR1A = 5; }
	TIMSK |= (1<<OCIE1A);										// Interrupt Timer erlauben
}

// ########## Interrupt Timer 1 ##########

ISR (TIMER1_COMPA_vect) {
	if (ucMaster) {
		// Sync auf 1. Frame
		if (iPos == 0 && ucSkipCounter == ucFirstSkipCounter) {
			if (ucColorCnt == 0 && ucRowCnt == 0) {
				PORTA = 0XFF;
				PORTB = 0XFF;
				for (ulWait=0; ulWait < 10000; ulWait++) {
					_NOP ();
				}
			}
		}
		Draw_Workspace ();
	}
	else {
		if ((PINC & 0x20) != ucLastSyncBit) {
			Draw_Workspace ();
			ucLastSyncBit ^= 0x20;
			iMissingSyncCounter = 0;
		}
		iMissingSyncCounter++;
		if (iMissingSyncCounter > 40) {
			iMissingSyncCounter = 41;
			iPos = 0;
			ucNewRow = 0;
			ucColorCnt = 0;
			ucRowCnt = 0;
			ucSkipCounter = ucFirstSkipCounter;
			ucFadeIn = 0;
			ucFadeOut = 0;
			PORTA = 0XFF;
			PORTB = 0XFF;
		}
	}
}

// ########## Den Arbeitsbereich aufs Bild bringen ##########
void Draw_Workspace () {
unsigned char ucTmpColorCnt;
unsigned char ucTmpColCnt;

	// Module Sync
	PORTC ^= 0x40;

	// USART Busy Counter
	if (uiUSARTBusyCounter > 0) {
		uiUSARTBusyCounter--;
		PORTA = 0XFF;
		PORTB = 0XFF;
		iPos = 0;
	}
	else {
		// Anzeige nur wenn keine USART Kommunikation stattfindet
		iUSARTBufferPointer = 0;
			
		ucColorCnt++;
		ucColorCnt &= 0x0F;
		if (ucColorCnt == 0) {
			ucRowCnt--;
			ucRowCnt &= 0x07;
			PORTC &= 0xE3;
			PORTC |= (ucRowCnt << 2);
			if (ucRowCnt == 7) {
				ucNewRow = 1;
				ucFrame++;
				ucFrame &= 0x01;	
				ucDrawFrame = (~ucFrame & 0x01);
			}

			for (ucColCnt=0; ucColCnt<8; ucColCnt++) {
				ucColor = 0x00;
				ucTmpColCnt = 1 << ucColCnt;
				if (ucWorkspace[ucFrame][0][ucRowCnt] & ucTmpColCnt) { ucColor |= 0x01; }
				if (ucWorkspace[ucFrame][1][ucRowCnt] & ucTmpColCnt) { ucColor |= 0x02; }
				if (ucWorkspace[ucFrame][2][ucRowCnt] & ucTmpColCnt) { ucColor |= 0x04; }
				switch (ucColor) {
					case 0:  red[ucColCnt] = 0x00; grn[ucColCnt] = 0x00; break;
					case 1:  red[ucColCnt] = 0x0F; grn[ucColCnt] = 0x00; break;
					case 2:  red[ucColCnt] = 0x0F; grn[ucColCnt] = 0x04; break;
					case 3:  red[ucColCnt] = 0x0F; grn[ucColCnt] = 0x08; break;
					case 4:  red[ucColCnt] = 0x0F; grn[ucColCnt] = 0x0F; break;
					case 5:  red[ucColCnt] = 0x04; grn[ucColCnt] = 0x0F; break;
					case 6:  red[ucColCnt] = 0x02; grn[ucColCnt] = 0x0F; break;
					case 7:  red[ucColCnt] = 0x00; grn[ucColCnt] = 0x0F; break;
				}
			}
		}
		
		ucTmpColorCnt = ucColorCnt + ucDark[ucFrame];
		ucPortA = 0x00;
		if (ucTmpColorCnt >= red[0]) { ucPortA |= 0x01; }
		if (ucTmpColorCnt >= red[1]) { ucPortA |= 0x02; }
		if (ucTmpColorCnt >= red[2]) { ucPortA |= 0x04; }
		if (ucTmpColorCnt >= red[3]) { ucPortA |= 0x08; }
		if (ucTmpColorCnt >= red[4]) { ucPortA |= 0x10; }
		if (ucTmpColorCnt >= red[5]) { ucPortA |= 0x20; }
		if (ucTmpColorCnt >= red[6]) { ucPortA |= 0x40; }
		if (ucTmpColorCnt >= red[7]) { ucPortA |= 0x80; }
		PORTA = ucPortA;
		ucPortB = 0x00;
		if (ucTmpColorCnt >= grn[0]) { ucPortB |= 0x01; }
		if (ucTmpColorCnt >= grn[1]) { ucPortB |= 0x02; }
		if (ucTmpColorCnt >= grn[2]) { ucPortB |= 0x04; }
		if (ucTmpColorCnt >= grn[3]) { ucPortB |= 0x08; }
		if (ucTmpColorCnt >= grn[4]) { ucPortB |= 0x10; }
		if (ucTmpColorCnt >= grn[5]) { ucPortB |= 0x20; }
		if (ucTmpColorCnt >= grn[6]) { ucPortB |= 0x40; }
		if (ucTmpColorCnt >= grn[7]) { ucPortB |= 0x80; }
		PORTB = ucPortB;
	}
}

// ########## I2C Init ##########
void I2C_Init (void) {
	TWBR = 2;										// SCL Freq = 16MHz/(16+2*TWBR*4^TWSR) = 800KHz
	TWSR = (0<<TWPS1)|(0<<TWPS0);					// Teiler = 0 
}

// ########## I2C Start ##########
void I2C_Start (void) {
	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);		// Send START condition
	Wait_TWINT_Flag ();
	if ((TWSR & 0xF8) != TW_START && (TWSR & 0xF8) != TW_REP_START) {	// Check value of TWI Status Register
		ucErr |= 0x01;
		TWCR = 0;
	}
}

// ########## I2C Stop ##########
void I2C_Stop (void) {
	TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);		// Transmit STOP condition
	Wait_TWSTO_Flag ();
}

// ########## I2C Adr Write ##########
void I2C_Adr_W (void) {
	TWDR = I2C_SLA_ADR << 1;
	TWCR = (1<<TWINT) | (1<<TWEN);					// Clear TWINT bit in TWCR to start transmission of address
	Wait_TWINT_Flag ();
	if ((TWSR & 0xF8) != TW_MT_SLA_ACK) {			// Check value of TWI Status Register
		ucErr |= 0x02;
		TWCR = 0;	
	}
}

// ########## I2C Adr Read ##########
void I2C_Adr_R (void) {
	TWDR = (I2C_SLA_ADR << 1) | 1;
	TWCR = (1<<TWINT) | (1<<TWEN);					// Clear TWINT bit in TWCR to start transmission of address
	Wait_TWINT_Flag ();
	if ((TWSR & 0xF8) != TW_MR_SLA_ACK) {			// Check value of TWI Status Register
		ucErr |= 0x04;
		TWCR = 0;
	}
}

// ########## I2C Data Write ##########
void I2C_Data_W (unsigned char ucByte) {
	TWDR = ucByte;
	TWCR = (1<<TWINT) | (1<<TWEN);					// Clear TWINT bit in TWCR to start transmission of address
	Wait_TWINT_Flag ();
	if ((TWSR & 0xF8) != TW_MT_DATA_ACK) {			// Check value of TWI Status Register
		ucErr |= 0x08;
		TWCR = 0;
	}
}

// ########## I2C Data Read NAck ##########
unsigned char I2C_Data_R (void) {
	TWCR = (1<<TWINT) | (1<<TWEN);					// Clear TWINT bit in TWCR to start transmission of address
	Wait_TWINT_Flag ();
	return TWDR;
}

// ########## I2C Data Read Ack ##########
unsigned char I2C_Data_R_Ack (void) {
	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);	// Clear TWINT bit in TWCR to start transmission of address
	Wait_TWINT_Flag ();
	return TWDR;
}

// ########## I2C Wait for TWINT Flag ##########
void Wait_TWINT_Flag (void) {
unsigned int uiCnt = 0;
	while (!(TWCR & (1<<TWINT))) {
		uiCnt++;
		if (uiCnt == 1000) {
			return;
		}
	}
}

// ########## I2C Wait for TWSTO Flag ##########
void Wait_TWSTO_Flag (void) {
unsigned int uiCnt = 0;
	while (!(TWCR & (1<<TWSTO))) {
		uiCnt++;
		if (uiCnt == 1000) {
			return;
		}
	}
}

// ########## Lese aus EEPROM ##########
unsigned char EE_Read (unsigned int iAdr) {
unsigned char ucLOByte;
unsigned char ucHIByte;
unsigned char ucByte;
	cli ();
	ucLOByte = iAdr;
	ucHIByte = iAdr >> 8;
	I2C_Start ();
	I2C_Adr_W ();
	I2C_Data_W (ucHIByte);
	I2C_Data_W (ucLOByte);
	I2C_Start ();
	I2C_Adr_R ();
	ucByte = I2C_Data_R ();
	I2C_Stop ();
	sei ();
	return ucByte;
}

// ########## Schreibe in EEPROM ##########
void EE_Write (unsigned int iAdr) {
unsigned char ucLOByte;
unsigned char ucHIByte;
unsigned ucPt;
	cli ();
	ucLOByte = iAdr;
	ucHIByte = iAdr >> 8;
	I2C_Start ();
	I2C_Adr_W ();
	I2C_Data_W (ucHIByte);
	I2C_Data_W (ucLOByte);
	for (ucPt=0; ucPt<128; ucPt++) {
		I2C_Data_W (ucEEWriteBuffer[ucPt]);
	}
	I2C_Stop ();
	sei ();
}

// ########## USART Init ##########
void USART_Init (void) {
unsigned int ubrr;
	ubrr = 0;										// ubrr = ( FREQ / ( 16 * 1000k) ) -1
	UBRRH = (unsigned char)(ubrr>>8);				// Baud-Rate
	UBRRL = (unsigned char)ubrr;
	UCSRA = (0<<U2X);								// USART 2x Speed
	UCSRB = (1<<RXEN)|(1<<TXEN);					// Enable Receiver und Transmitter
	UCSRC = (1<<URSEL)|(3<<UCSZ0);					// Format setzen: 8 Data, 1 Stop bit, kein Parity
}

// ########## USART Transmit ##########
void USART_Transmit (char *data) {
int i = 0;
	while (data[i] != '\0') {
		while ( !( UCSRA & (1<<UDRE)) ) {;}		// Warte auf leeren Transmit Buffer
		UDR = data[i];								// Daten zum Puffer und Daten senden
		i++;
	}
}

// ########## USART Receive ##########
void USART_Receive (void) {
int i = 0;
char cData;
	while ( !(UCSRA & (1<<RXC)) && i < 10 ) {		// Warte auf Daten zum Empfangen
		i++;
	}
	if (i < 10) {
		cData = UDR;
		uiUSARTBusyCounter = 32768;
		if (cData == '' && iUSARTBufferPointer == 0) {	// Start empfangen
			sUSARTBuffer[iUSARTBufferPointer] = cData;
			iUSARTBufferPointer++;
		}
		else {
			if ( iUSARTBufferPointer > 0) {
				sUSARTBuffer[iUSARTBufferPointer] = cData;
				iUSARTBufferPointer++;
				if (iUSARTBufferPointer > 135) {
					iUSARTBufferPointer = 0;
				}
			}
		}
	}
}

// ########## Execute Command ##########
void USART_Exec_Commands (void) {
unsigned char ucIdx;
unsigned char ucChk;
unsigned int iWrAdr;
	if (iUSARTBufferPointer > 2) {
		// Command Identify Device
		if (sUSARTBuffer[1] == 'I' && sUSARTBuffer[2] == 'D') {
			USART_Transmit("PROMOD\r\n");
			sUSARTBuffer[1] = 0;
			sUSARTBuffer[2] = 0;
			iUSARTBufferPointer = 0;
		}
		// Command WRite data
		if (sUSARTBuffer[1] == 'W' && sUSARTBuffer[2] == 'R' && sUSARTBuffer[134] == '#') {
			ucChk = 0xAA;
			iWrAdr = sUSARTBuffer[3] << 8;
			ucChk -= sUSARTBuffer[3];
			iWrAdr |= sUSARTBuffer[4];
			ucChk -= sUSARTBuffer[4];
			for (ucIdx=0; ucIdx<128; ucIdx++) {
				ucEEWriteBuffer[ucIdx] = (sUSARTBuffer[ucIdx+5]);
				ucChk -= ucEEWriteBuffer[ucIdx];
			}
			sUSARTBuffer[1] = 0;
			sUSARTBuffer[2] = 0;
			sUSARTBuffer[134] = 0;
			iUSARTBufferPointer = 0;
			if (sUSARTBuffer[133]!= ucChk) {
				USART_Transmit("ERR\r\n");
				ucErr |= 0x10;
				return;
			}
			USART_Transmit("OK\r\n");
			EE_Write (iWrAdr << 7);
		}
	}
}
